using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Threading.Tasks;
using FarmmapsApi.Models;
using FarmmapsApi.Services;
using Microsoft.Extensions.Logging;
using Newtonsoft.Json.Linq;
using static FarmmapsApi.Extensions;
using static FarmmapsApiSamples.Constants;

namespace FarmmapsApiSamples
{
    public class NitrogenService : ITestFlow
    {
        private readonly ILogger<NitrogenService> _logger;
        private readonly FarmmapsApiService _farmmapsApiService;
        private readonly GeneralService _generalService;

        public NitrogenService(ILogger<NitrogenService> logger, FarmmapsApiService farmmapsApiService,
            GeneralService generalService)
        {
            _logger = logger;
            _farmmapsApiService = farmmapsApiService;
            _generalService = generalService;
        }

        public async Task TestFlow(List<UserRoot> roots)
        {
            var downloadFolder = "Downloads";
            if (!Directory.Exists(downloadFolder))
                Directory.CreateDirectory(downloadFolder);
            
            var plantingDate = new DateTime(DateTime.UtcNow.Year, 2, 1);
            var measurementDate = plantingDate.AddMonths(5);
            
            var uploadedRoot = roots.SingleOrDefault(r => r.Name == "Uploaded");
            if (uploadedRoot == null)
            {
                _logger.LogError("Could not find a needed root item");
                return;
            }
            
            var myDriveRoot = roots.SingleOrDefault(r => r.Name == "My drive");
            if (myDriveRoot == null)
            {
                _logger.LogError("Could not find a needed root item");
                return;
            }
            
            var cropfieldItem = await GetOrCreateCropfieldItemAsync(myDriveRoot.Code);
            
            var dataPath = Path.Combine("Data", "Scan_1_20190605.zip");
            var isariaShapeItem = await _generalService.UploadZipWithShapeAsync(uploadedRoot, dataPath, "Scan_1_20190605");
            if (isariaShapeItem == null)
            {
                _logger.LogError("Could not find isaria shape item");
                return;
            }
            
            var isariaGeotiffItem = await _generalService.ShapeToGeotiff(isariaShapeItem);
            if (isariaGeotiffItem == null)
            {
                _logger.LogError("Something went wrong with isaria shape to geotiff transformation");
                return;
            }

            _logger.LogInformation($"Calculating targetN with targetYield: {60}");
            var targetN = await CalculateTargetN(cropfieldItem, plantingDate, measurementDate, 60);
            _logger.LogInformation($"TargetN: {targetN}");
            
            _logger.LogInformation("Calculating uptake map");
            var uptakeMapItem =
                await CalculateUptakeMap(cropfieldItem, isariaGeotiffItem, plantingDate, measurementDate);
            
            _logger.LogInformation("Downloading uptake map");
            await _farmmapsApiService.DownloadItemAsync(uptakeMapItem.Code, Path.Combine(downloadFolder, $"{uptakeMapItem.Name}.zip"));

            _logger.LogInformation("Calculating appliance map");
            var applianceMapItem =
                await CalculateApplianceMap(cropfieldItem, isariaGeotiffItem, plantingDate, measurementDate, targetN);

            _logger.LogInformation("Downloading appliance map");
            await _farmmapsApiService.DownloadItemAsync(applianceMapItem.Code, Path.Combine(downloadFolder, $"{applianceMapItem.Name}.zip"));
        }
        
        private async Task<Item> GetOrCreateCropfieldItemAsync(string parentItemCode)
        {
            var cropfieldItems = await
                _farmmapsApiService.GetItemChildrenAsync(parentItemCode, CROPFIELD_ITEMTYPE);

            if (cropfieldItems.Count > 0)
                return cropfieldItems[0];

            var currentYear = new DateTime(DateTime.UtcNow.Year, 1, 1);
            var cropfieldItemRequest = new ItemRequest()
            {
                ParentCode = parentItemCode,
                ItemType = CROPFIELD_ITEMTYPE,
                Name = "Cropfield for VRA",
                DataDate = currentYear,
                DataEndDate = currentYear.AddYears(1).AddDays(-1),
                Data = JObject.Parse("{}"),
                Geometry = JObject.Parse(
                    @"{ ""type"": ""Polygon"", ""coordinates"": [ [ [ 3.40843828875524, 50.638966444680605 ], [ 3.408953272886064, 50.639197789621612 ], [ 3.409242951459603, 50.639469958681836 ], [ 3.409328782148028, 50.639612846807708 ], [ 3.409457528180712, 50.639789755314411 ], [ 3.409639918393741, 50.640014292074966 ], [ 3.409833037442765, 50.640211611372706 ], [ 3.410069071836049, 50.640395321698435 ], [ 3.410380208081761, 50.640572227259661 ], [ 3.410605513638958, 50.640715112034222 ], [ 3.411925160474145, 50.641177783561204 ], [ 3.411935889310142, 50.640728720085136 ], [ 3.412590348309737, 50.63948356709389 ], [ 3.413244807309242, 50.638224772339846 ], [ 3.413400375432099, 50.637901562841307 ], [ 3.413539850300779, 50.637449065809889 ], [ 3.413475477284437, 50.637418445552932 ], [ 3.40999396998362, 50.637449065810451 ], [ 3.409940325803365, 50.638102293212661 ], [ 3.409575545377398, 50.638483338338325 ], [ 3.409060561246574, 50.638707881340494 ], [ 3.40843828875524, 50.638966444680605 ] ] ] }")
            };

            return await _farmmapsApiService.CreateItemAsync(cropfieldItemRequest);
        }
        
        /// <summary>
        /// Calculates TargetN, makes the assumption the cropfield and user.input(targetn) item have the same parent
        /// </summary>
        /// <param name="cropfieldItem">The cropfield to base the calculations on</param>
        /// <param name="targetYield">The target yield input for the TargetN calculation</param>
        /// <param name="plantingDate">The date the crop is planted</param>
        /// <param name="measurementDate">The date the measurements are taken</param>
        /// <returns>The TargetN</returns>
        public async Task<double> CalculateTargetN(Item cropfieldItem, DateTime plantingDate, 
            DateTime measurementDate, int targetYield)
        {
            var targetNItems = await
                _farmmapsApiService.GetItemChildrenAsync(cropfieldItem.ParentCode, USERINPUT_ITEMTYPE);

            Item targetNItem;
            if (targetNItems.Count == 0)
            {
                _logger.LogInformation("Creating targetN item");
                var itemRequest = CreateTargetNItemRequest(cropfieldItem.ParentCode);
                targetNItem = await _farmmapsApiService.CreateItemAsync(itemRequest);
            }
            else
            {
                targetNItem = targetNItems[0];
            }

            var nbsTargetNRequest = new TaskRequest {TaskType = VRANBS_TASK};
            nbsTargetNRequest.attributes["operation"] = "targetn";
            nbsTargetNRequest.attributes["inputCode"] = targetNItem.Code;
            nbsTargetNRequest.attributes["plantingDate"] = plantingDate.ToString();
            nbsTargetNRequest.attributes["measurementDate"] = measurementDate.ToString();
            nbsTargetNRequest.attributes["inputType"] = "irmi";
            nbsTargetNRequest.attributes["purposeType"] = "consumption";
            nbsTargetNRequest.attributes["targetYield"] = targetYield.ToString();
            string itemTaskCode = await _farmmapsApiService.QueueTaskAsync(cropfieldItem.Code, nbsTargetNRequest);
            
            await PollTask(TimeSpan.FromSeconds(3), async (tokenSource) =>
            {
                var itemTaskStatus = await _farmmapsApiService.GetTaskStatusAsync(cropfieldItem.Code, itemTaskCode);
                if (itemTaskStatus.IsFinished)
                    tokenSource.Cancel();
            });

            var itemTask = await _farmmapsApiService.GetTaskStatusAsync(cropfieldItem.Code, itemTaskCode);
            if(itemTask.State == ItemTaskState.Error)
            {
                _logger.LogError($"Something went wrong with task execution: {itemTask.Message}");
                return 0;
            }
            
            var item = await _farmmapsApiService.GetItemAsync(targetNItem.Code);
            if (item.Data.ContainsKey("TargetN"))
                return item.Data.Value<double>("TargetN");

            return 0;
        }
        
        /// <summary>
        /// Calculates the uptake map based on the given inputs
        /// </summary>
        /// <param name="cropfieldItem">The cropfield to base the calculations on</param>
        /// <param name="inputItem"></param>
        /// <param name="plantingDate">The date the crop is planted</param>
        /// <param name="measurementDate">The date the measurements are taken</param>
        /// <returns></returns>
        public async Task<Item> CalculateUptakeMap(Item cropfieldItem, Item inputItem, DateTime plantingDate, 
            DateTime measurementDate)
        {
            var nbsUptakeMapRequest = new TaskRequest {TaskType = VRANBS_TASK};
            nbsUptakeMapRequest.attributes["operation"] = "uptake";
            nbsUptakeMapRequest.attributes["inputCode"] = inputItem.Code;
            nbsUptakeMapRequest.attributes["plantingDate"] = plantingDate.ToString();
            nbsUptakeMapRequest.attributes["measurementDate"] = measurementDate.ToString();
            nbsUptakeMapRequest.attributes["inputType"] = "irmi";

            string itemTaskCode = await _farmmapsApiService.QueueTaskAsync(cropfieldItem.Code, nbsUptakeMapRequest);
            
            await PollTask(TimeSpan.FromSeconds(5), async (tokenSource) =>
            {
                var itemTaskStatus = await _farmmapsApiService.GetTaskStatusAsync(cropfieldItem.Code, itemTaskCode);
                if (itemTaskStatus.IsFinished)
                    tokenSource.Cancel();
            });
            
            var itemTask = await _farmmapsApiService.GetTaskStatusAsync(cropfieldItem.Code, itemTaskCode);
            if(itemTask.State == ItemTaskState.Error)
            {
                _logger.LogError($"Something went wrong with task execution: {itemTask.Message}");
                return null;
            }
            
            var geotiffItems = await
                _farmmapsApiService.GetItemChildrenAsync(cropfieldItem.Code, GEOTIFF_PROCESSED_ITEMTYPE);
            
            // maybe poll here to, to wait for geotiff task process to be finished?

            // how to uniquely know which item is really created!?
            var nitrogenItem = geotiffItems.SingleOrDefault(i => i.Name.Contains("uptake") && i.Updated >= itemTask.Finished);
            if (nitrogenItem == null)
            {
                _logger.LogError("Could not find the uptake geotiff child item under cropfield");
                return null;
            }

            return nitrogenItem;
        }


        /// <summary>
        /// Creates the nitrogen appliance map based on given input data
        /// </summary>
        /// <param name="cropfieldItem">The cropfield to base the calculations on</param>
        /// <param name="inputItem">The farmmaps item containing the geotiff data</param>
        /// <param name="plantingDate">The date the crop is planted</param>
        /// <param name="measurementDate">The date the measurements are taken</param>
        /// <param name="targetN">The target nitrogen to use for the calculations</param>
        /// <returns></returns>
        public async Task<Item> CalculateApplianceMap(Item cropfieldItem, Item inputItem, DateTime plantingDate, 
            DateTime measurementDate, double targetN)
        {
            var nbsApplianceMapRequest = new TaskRequest {TaskType = VRANBS_TASK};
            nbsApplianceMapRequest.attributes["operation"] = "nitrogen";
            nbsApplianceMapRequest.attributes["inputCode"] = inputItem.Code;
            nbsApplianceMapRequest.attributes["plantingDate"] = plantingDate.ToString();
            nbsApplianceMapRequest.attributes["measurementDate"] = measurementDate.ToString();
            nbsApplianceMapRequest.attributes["inputCode"] = inputItem.Code;
            nbsApplianceMapRequest.attributes["inputType"] = "irmi";
            nbsApplianceMapRequest.attributes["targetN"] = targetN.ToString(CultureInfo.InvariantCulture);
            
            string itemTaskCode = await _farmmapsApiService.QueueTaskAsync(cropfieldItem.Code, nbsApplianceMapRequest);
            
            await PollTask(TimeSpan.FromSeconds(5), async (tokenSource) =>
            {
                var itemTaskStatus = await _farmmapsApiService.GetTaskStatusAsync(cropfieldItem.Code, itemTaskCode);
                if (itemTaskStatus.IsFinished)
                    tokenSource.Cancel();
            });
            
            var itemTask = await _farmmapsApiService.GetTaskStatusAsync(cropfieldItem.Code, itemTaskCode);
            if(itemTask.State == ItemTaskState.Error)
            {
                _logger.LogError($"Something went wrong with task execution: {itemTask.Message}");
                return null;
            }
            
            var geotiffItems = await
                _farmmapsApiService.GetItemChildrenAsync(cropfieldItem.Code, GEOTIFF_PROCESSED_ITEMTYPE);
            
            // maybe poll here too, to wait for geotiff task process to be finished?

            // how to uniquely know which item is really created!?
            var applianceMapItem = geotiffItems.SingleOrDefault(i => i.Name.Contains("nitrogen") && i.Updated >= itemTask.Finished);
            if (applianceMapItem == null)
            {
                _logger.LogError("Could not find the nitrogen geotiff child item under cropfield");
                return null;
            }

            return applianceMapItem;
        }

        private ItemRequest CreateTargetNItemRequest(string parentItemCode)
        {
            return new ItemRequest()
            {
                ParentCode = parentItemCode,
                ItemType = USERINPUT_ITEMTYPE,
                Name = "TargetN"
            };
        }
    }
}